@@ -1,5 +1,6 @@
 /* ====================================================================
- * Copyright (c) 1995, 1996, 1997, 1998, 1999 Ben Laurie.  All rights reserved.
+ * Copyright (c) 1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002 Ben Laurie.
+ *  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -77,10 +78,17 @@
 #define RENEG		FALSE
 #define DEBUG_RENEG	FALSE
 #define DEBUG_SSL_STATE FALSE
+#define ENGINE_SUPPORT	FALSE
+
+#if ENGINE_SUPPORT
+#include <openssl/engine.h>
+#endif
 
 #define USE_OLD_ENVIRONMENT	FALSE	/* Set to TRUE to use the old SSL_{CLIENT,SERVER}_I* instead of SSL_{CLIENT,SERVER}_I_* */
 
-#define SSL_IS_OPTIONAL		FALSE	/* Set to TRUE if you want to allow SSL to be disabled on a per-request basis (useful for subrequests) */
+#define SSL_IS_OPTIONAL		TRUE	/* Set to TRUE if you want to allow SSL to be disabled on a per-request basis (useful for subrequests) */
+
+#define USE_EXTENDED_ENVIRONMENT      TRUE    /* Set to TRUE to use the SSL_{CLIENT,SERVER}_{I,S}_DN* */
 
 #if SSLEAY_VERSION_NUMBER >= 0x0800
 /* Do session caching - only available with SSLeay 0.8.x */
@@ -112,11 +120,17 @@
 static const char *s_szCacheServerPort;
 static const char *s_szCacheServerDir;
 static char *s_szCacheServerPath;
+static char *s_szCacheServerArg1;
+static char *s_szCacheServerArg2;
 
 # define DEBUG_SESSIONS	FALSE
 
 #endif
 
+#if ENGINE_SUPPORT
+static const char *s_szEngineID = NULL;
+#endif
+
 #if APACHE_SSL_EXPORT_CERTS && !CACHE_SESSIONS
 # error "CACHE_SESSIONS must be TRUE if APACHE_SSL_EXPORT_CERTS is"
 #endif
@@ -185,6 +199,7 @@ typedef struct
     char *szCACertificatePath;
     char *szCACertificateFile;
     char *szReqCiphers;
+    char *szDNCheckFile;
 #if !RENEG
     int nVerifyDepth;
     VerifyType nVerifyClient;
@@ -311,8 +326,8 @@ static void InfoCallback(SSL *s,int where,int ret)
 	str="undefined";
 
     if(where & SSL_CB_LOOP)
-	ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,"%s:%s",
-		     str,SSL_state_string_long(s));
+	ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,"%s:%s (%s)",
+		     str,SSL_state_string_long(s),SSL_get_cipher(s));
     else if(where & SSL_CB_ALERT)
 	{
 	str=(where & SSL_CB_READ)?"read":"write";
@@ -339,8 +354,15 @@ static void InfoCallback(SSL *s,int where,int ret)
 
 #endif /* DEBUG_SSL_STATE */
 
+typedef enum
+    {
+    RAND_EGD,
+    RAND_FILE
+    } RFType;
+
 typedef struct
     {
+    RFType eType;
     const char *szFile;
     int nBytes;
     } RandomFile;
@@ -351,7 +373,9 @@ static void ReadRandomFiles(server_rec *pServer,RandomFile *pRF,int nRF)
 
     for(n=0 ; n < nRF ; ++n,++pRF)
 	{
-	int r=RAND_load_file(pRF->szFile,pRF->nBytes);
+	int r=pRF->eType == RAND_EGD
+	  ? RAND_egd_bytes(pRF->szFile,pRF->nBytes)
+	  : RAND_load_file(pRF->szFile,pRF->nBytes);
 	ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,
 		     "Random input %s(%d) -> %d",pRF->szFile,pRF->nBytes,r);
 	}
@@ -436,7 +460,11 @@ static int ApacheSSLVerifyCallback(int ok,X509 *xs,X509 *xi,int depth,
     if(error == X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY
        || error == X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE
        || error == X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT
-       || error == X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN)
+       || error == X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN
+#if OPENSSL_VERSION_NUMBER >= 0x00905000L
+       || error == X509_V_ERR_CERT_UNTRUSTED
+#endif
+       )
 	{
 	char *s;
 
@@ -528,6 +556,8 @@ static int SSLCheckCipher(request_rec *r)
     if(pConfig->bDisabled || OPTIONAL_SSL)
 	return DECLINED;
 
+    if (r->connection->client->ssl == NULL) return DECLINED;
+
 #if RENEG
 # if DEBUG_RENEG
     fprintf(stderr,"Renegotiated for %s (%x)!\n",r->filename,VerifyFlags(rec));
@@ -598,13 +628,111 @@ static int SSLCheckCipher(request_rec *r)
 	}
     }
 
+static int CheckDN(request_rec *r)
+    {
+    SSLConfigRec *pConfig=ap_get_module_config(r->server->module_config,
+					       &apache_ssl_module);
+    FILE *f;
+  
+    if(OPTIONAL_SSL)
+	return DECLINED;
+    if(!pConfig->szDNCheckFile)
+	return DECLINED;
+    if(r->connection->user)
+	return DECLINED;
+    if(!r->connection->client->szClientX509)
+	return DECLINED;
+
+    f=fopen(pConfig->szDNCheckFile,"r");
+    if(!f)
+        {
+	ap_log_error(APLOG_MARK,APLOG_CRIT,r->server,
+		     "Can't open DN check file: %s",pConfig->szDNCheckFile);
+	return FORBIDDEN;
+	}
+
+    for( ; ; )
+        {
+	char buf[MAX_STRING_LEN];
+	int n;
+
+	if(!fgets(buf,sizeof buf,f))
+	    break;
+	n=strlen(buf);
+	if(n && buf[n-1] == '\n')
+	    buf[n-1]='\0';
+
+	ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,r->server,
+		     "Comparing '%s' to '%s'",buf,
+		     r->connection->client->szClientX509);
+
+	if(!strcmp(buf,r->connection->client->szClientX509))
+	    {
+	    fclose(f);
+	    return OK;
+	    }
+	}
+    fclose(f);
+
+    ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,r->server,
+		 "DN '%s' not found in %s",r->connection->client->szClientX509,
+		 pConfig->szDNCheckFile);
+
+    return FORBIDDEN;
+    }
+
+static int CheckAccess(request_rec *r)
+    {
+    int ret;
+
+    ret=SSLCheckCipher(r);
+    if(ret != OK)
+        return ret;
+    return CheckDN(r);
+    }
+
 static void StrUpper(char *s)
     {
     for( ; *s ; ++s)
 	*s=toupper(*s);
     }
 
-static void ExpandCert(pool *p,table *pEnv,char *szPrefix,char *szCert)
+#if USE_EXTENDED_ENVIRONMENT
+static void get_serial(pool *p,table *pEnv,char *name, X509 *xs)
+{
+    BIO *bio;
+    char *result;
+    int n;
+
+    if ((bio = BIO_new(BIO_s_mem())) == NULL) return;
+    i2a_ASN1_INTEGER(bio, X509_get_serialNumber(xs));
+    n = BIO_pending(bio);
+    result = ap_pcalloc(p, n+1);
+    n = BIO_read(bio, result, n);
+    result[n] = 0;
+    ap_table_setn(pEnv,name, result);
+    BIO_free(bio);
+}
+
+static void get_valid(pool *p, table *pEnv, char *name, ASN1_UTCTIME *tm)
+{
+    BIO *bio;
+    char *result;
+    int n;
+
+    if ((bio = BIO_new(BIO_s_mem())) == NULL) return;
+    ASN1_UTCTIME_print(bio, tm);
+    n = BIO_pending(bio);
+    result = ap_pcalloc(p, n+1);
+    n = BIO_read(bio, result, n);
+    result[n] = 0;
+    ap_table_setn(pEnv, name, result);
+    BIO_free(bio);
+}
+
+#endif
+
+static void ExpandCert(pool *p,table *pEnv,char *szPrefix, char *szDN, char *szCert)
     {
     char buf[HUGE_STRING_LEN];
     char *s,*t;
@@ -614,8 +742,7 @@ static void ExpandCert(pool *p,table *pEnv,char *szPrefix,char *szCert)
        with CGI's written for stronghold's certificate info  - Q */
     /* FIXME - strtok() and strcspn() may cause problems on some systems - Q */
 
-    t=ap_psprintf(p,"%sDN",szPrefix);
-    ap_table_setn(pEnv,t,ap_pstrdup(p,szCert));
+    ap_table_setn(pEnv,szDN,ap_pstrdup(p,szCert));
     
     ap_cpystrn(buf,szCert,sizeof buf);
     for(s=strtok(buf,"/") ; s != NULL ; s=strtok(NULL,"/"))
@@ -677,7 +804,7 @@ static struct
     { NULL, 0, 0 }
     };
 
-void ApacheSSLAddCommonVars(request_rec *r)
+static void SSLAddCommonVars(request_rec *r)
     {
     table *e=r->subprocess_env;
     int keysize=0;
@@ -701,6 +828,7 @@ void ApacheSSLAddCommonVars(request_rec *r)
     if(pConfig->bDisabled || OPTIONAL_SSL)
 	return;
 
+    if (r->connection->client->ssl == NULL) return;
     cipher=SSL_get_cipher(r->connection->client->ssl);
     ap_table_set(e,"HTTPS","on");
     ap_table_set(e,"HTTPS_CIPHER",cipher);
@@ -725,16 +853,29 @@ void ApacheSSLAddCommonVars(request_rec *r)
 
     if(r->connection->client->szClientX509)
 	{
-	ExpandCert(r->pool,e,"SSL_CLIENT_",
+	ExpandCert(r->pool,e,"SSL_CLIENT_","SSL_CLIENT_DN",
 		   r->connection->client->szClientX509);
+#if USE_EXTENDED_ENVIRONMENT
+	ExpandCert(r->pool,e,"SSL_CLIENT_S_DN_","SSL_CLIENT_S_DN",
+		   r->connection->client->szClientX509);
+#endif
 	xs=SSL_get_peer_certificate(r->connection->client->ssl);
 	ExpandCert(r->pool,e,
 #if USE_OLD_ENVIRONMENT
-		   "SSL_CLIENT_I",
+		   "SSL_CLIENT_I","SSL_CLIENT_I_DN",
 #else
-		   "SSL_CLIENT_I_",
+		   "SSL_CLIENT_I_","SSL_CLIENT_I_DN",
 #endif
 		   X509_NAME_oneline(X509_get_issuer_name(xs),NULL,0));
+#if USE_EXTENDED_ENVIRONMENT
+	ExpandCert(r->pool,e,"SSL_CLIENT_I_DN_","SSL_CLIENT_I_DN",
+		   X509_NAME_oneline(X509_get_issuer_name(xs),NULL,0));
+	get_serial(r->pool,e,"SSL_CLIENT_M_SERIAL",xs);
+        get_valid(r->pool,e,"SSL_CLIENT_V_START",X509_get_notBefore(xs));
+        get_valid(r->pool,e,"SSL_CLIENT_V_END",X509_get_notAfter(xs));
+        t=ap_psprintf(r->pool,"%lu",X509_get_version(xs)+1);
+        ap_table_setn(e,"SSL_CLIENT_M_VERSION",t);
+#endif
 #if APACHE_SSL_EXPORT_CERTS
 	if(pDirConfig->bExportCerts)
 	    AddCertToEnv(r->pool,e,"SSL_CLIENT_CERT",xs);
@@ -742,15 +883,29 @@ void ApacheSSLAddCommonVars(request_rec *r)
 	}
 
     xs=SSL_get_certificate(r->connection->client->ssl);
-    ExpandCert(r->pool,e,"SSL_SERVER_",
+    ExpandCert(r->pool,e,"SSL_SERVER_","SSL_SERVER_DN",
 	       X509_NAME_oneline(X509_get_subject_name(xs),NULL,0));
+#if USE_EXTENDED_ENVIRONMENT
+    ExpandCert(r->pool,e,"SSL_SERVER_S_","SSL_SERVER_S_DN",
+               X509_NAME_oneline(X509_get_subject_name(xs),NULL,0));
+#endif
     ExpandCert(r->pool,e,
 #if USE_OLD_ENVIRONMENT
-	       "SSL_SERVER_I",
+	       "SSL_SERVER_I","SSL_SERVER_I_DN",
 #else
-	       "SSL_SERVER_I_",
+	       "SSL_SERVER_I_","SSL_SERVER_I_DN",
 #endif
 	       X509_NAME_oneline(X509_get_issuer_name(xs),NULL,0));
+#if USE_EXTENDED_ENVIRONMENT
+    ExpandCert(r->pool,e,
+	       "SSL_SERVER_I_DN_","SSL_SERVER_I_DN",
+	       X509_NAME_oneline(X509_get_issuer_name(xs),NULL,0));
+    get_serial(r->pool,e,"SSL_SERVER_M_SERIAL",xs);
+    get_valid(r->pool,e,"SSL_SERVER_V_START",X509_get_notBefore(xs));
+    get_valid(r->pool,e,"SSL_SERVER_V_END",X509_get_notAfter(xs));
+    t=ap_psprintf(r->pool,"%lu",X509_get_version(xs)+1);
+    ap_table_setn(e,"SSL_SERVER_M_VERSION",t);
+#endif
 
 #if APACHE_SSL_EXPORT_CERTS
 # if !APACHE_SSL_KEEP_CERTS
@@ -789,6 +944,7 @@ static void *CreateSSLServerConfig(pool *p,server_rec *s)
     rec->szCACertificatePath=NULL;
     rec->szCACertificateFile=NULL;
     rec->szReqCiphers=NULL;
+    rec->szDNCheckFile=NULL;
 #if !RENEG
     rec->nVerifyDepth=0;
     rec->nVerifyClient=VERIFY_UNSET;	/* Will become VERIFY_NONE */
@@ -820,6 +976,8 @@ static void *MergeSSLServerConfig(pool *p,void *basev,void *addv)
     MergeString(szKeyFile);
     MergeString(szCACertificatePath);
     MergeString(szCACertificateFile);
+    MergeString(szDNCheckFile);
+    MergeString(szReqCiphers);
 #if !RENEG
     MergeInt(nVerifyDepth);
     Merge(nVerifyClient,VERIFY_UNSET);
@@ -897,6 +1055,44 @@ static void InitSSL(server_rec *s)
 #else
     /* only enable this with care: it can be used for /tmp symlink attacks.
     SSL_debug("/tmp/ssldebug");*/
+#endif
+#if ENGINE_SUPPORT
+    if(s_szEngineID)
+        {
+        ENGINE *e = ENGINE_by_id(s_szEngineID);
+        if(!e)
+            {
+            ap_log_error(APLOG_MARK,APLOG_CRIT|APLOG_NOERRNO,s,
+                         "Supplied SSLEngineID is unrecognised");
+            exit(1);
+            }
+        /* Squeeze in a quick hack - nCipher "chil" units don't do this by
+         * default, which means that child processes could inherent the same
+         * file-descriptor and corrupt each other's I/O. :-( */
+        if(strcmp(s_szEngineID, "chil"))
+            ENGINE_ctrl(e, ENGINE_CTRL_CHIL_SET_FORKCHECK, 1, 0, 0);
+        if(!ENGINE_set_default(e, ENGINE_METHOD_ALL))
+            {
+            ap_log_error(APLOG_MARK,APLOG_CRIT|APLOG_NOERRNO,s,
+                         "Supplied SSLEngineID could not be initialised");
+            exit(1);
+            }
+        ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,s,
+                     "SSLEngineID (as supplied) initialised and running");
+        if(!ENGINE_get_RSA(e))
+            ap_log_error(APLOG_MARK,APLOG_WARNING|APLOG_NOERRNO,s,
+                         "Supplied SSLEngineID does not support RSA");
+        else
+            ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,s,
+                         "Supplied SSLEngineID supports RSA");
+        if(!ENGINE_get_DSA(e))
+            ap_log_error(APLOG_MARK,APLOG_WARNING|APLOG_NOERRNO,s,
+                         "Supplied SSLEngineID does not support DSA");
+        else
+            ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,s,
+                         "Supplied SSLEngineID supports DSA");
+        ENGINE_free(e);
+        }
 #endif
     }
 
@@ -965,6 +1161,9 @@ static uchar *CacheFind(uchar *aucKey,int nKey,int *pnData,BOOL *pbLocal)
 	t=GlobalCacheGet(aucKey,nKey,pnData,&tExpiresAt);
 	if(!t)
 	    return NULL;
+#if DEBUG_SESSIONS
+	fprintf(stderr,"Session found in global cache, to=%ld\n", tExpiresAt);
+#endif
 
 	t2=malloc(nKey);
 	memcpy(t2,aucKey,nKey);
@@ -977,6 +1176,9 @@ static uchar *CacheFind(uchar *aucKey,int nKey,int *pnData,BOOL *pbLocal)
 	t=p->aucData;
 	*pnData=p->nData;
 	*pbLocal=TRUE;
+#if DEBUG_SESSIONS
+	fprintf(stderr,"Session found in local cache\n");
+#endif
 	}
     return t;
     }
@@ -1052,7 +1254,6 @@ static void GetCertChainFromServer(uchar *aucSessionID,int nLength,
 static void SendSessionToServer(SSL_SESSION *pSession,
 				conn_rec *pCurrentConnection)
     {
-    uchar buf[MAX_SESSION_DER];
     Cache *p;
     uchar *t;
     int nLength;
@@ -1065,18 +1266,19 @@ static void SendSessionToServer(SSL_SESSION *pSession,
     pConfig=ap_get_module_config(pCurrentConnection->server->module_config,
 				 &apache_ssl_module);
     tExpiresAt=time(NULL)+pConfig->tSessionCacheTimeout;
+    SSL_set_timeout(pSession, pConfig->tSessionCacheTimeout);
 
     t=malloc(pSession->session_id_length);
     memcpy(t,pSession->session_id,pSession->session_id_length);
     p=LocalCacheAdd(t,pSession->session_id_length,tExpiresAt);
 
-    t=buf;
-    nLength=i2d_SSL_SESSION(pSession,&t);
+    nLength=i2d_SSL_SESSION(pSession,NULL);
     assert(nLength > 0);
     assert(nLength <= MAX_SESSION_DER);
 
     p->aucData=malloc(nLength);
-    memcpy(p->aucData,buf,nLength);
+    t=p->aucData;
+    nLength=i2d_SSL_SESSION(pSession,&t);
     p->nData=nLength;
 
     GlobalCacheAdd(p->aucKey,p->nKey,p->aucData,p->nData,tExpiresAt);
@@ -1099,7 +1301,12 @@ static SSL_SESSION *GetSessionFromServer(uchar *aucSessionID,int nLength,
 
     t=CacheFind(aucSessionID,nLength,&nData,&bLocal);
     if(!t)
+	{
+#if DEBUG_SESSIONS
+	fprintf(stderr,"Session not found\n");
+#endif
 	return NULL;
+	}
 
     pSession=d2i_SSL_SESSION(NULL,&t,nData);
     if(!pSession)
@@ -1403,7 +1610,8 @@ static int CacheLauncher(void *cmd,child_info *pInfo)
 	}
     ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,s_pServer,
 		 "gcache pid is %d",getpid());
-    execl(cmd,"gcache",a1,s_szCacheServerPort,NULL);
+    execl(cmd,cmd,a1,s_szCacheServerPort,s_szCacheServerArg1,
+	  s_szCacheServerArg2,NULL);
     perror(cmd);
     assert(0);
     return 1;
@@ -1623,7 +1831,7 @@ static void CloseSSL(void *_conn)
     }
 #endif
 
-int ApacheSSLSetupConnection(conn_rec *conn)
+static int SSLSetupConnection(conn_rec *conn)
     {
     server_rec *srvr=conn->server;
     BUFF *fb=conn->client;
@@ -1727,6 +1935,9 @@ int ApacheSSLSetupConnection(conn_rec *conn)
     return TRUE;
     }
 
+static void SSLSetupConnection2(conn_rec *conn)
+    { SSLSetupConnection(conn); }
+
 static const char *SSLLogFile(cmd_parms *cmd,char *struct_ptr,char *arg)
     {
     return "The SSLLogFile directive is no longer used. All logging information now goes to the error log.";
@@ -1794,9 +2005,11 @@ static const char *SSLCacheServerPort(cmd_parms *cmd,char *struct_ptr,
     }
 
 static const char *SSLCacheServerPath(cmd_parms *cmd,char *struct_ptr,
-				      char *arg)
+				      char *arg,char *arg2,char *arg3)
     {
     s_szCacheServerPath=ap_server_root_relative(cmd->pool,arg);
+    s_szCacheServerArg1=arg2;
+    s_szCacheServerArg2=arg3;
 
     return NULL;
     }
@@ -1810,27 +2023,47 @@ static const char *SSLCacheServerDir(cmd_parms *cmd,char *struct_ptr,
     }
 #endif
 
+#if ENGINE_SUPPORT
+static const char *SSLEngineID(cmd_parms *cmd,char *struct_ptr,char *arg)
+    {
+    s_szEngineID=arg;
+    return NULL;
+    }
+#endif
+
+static void SetupRandomFile(RandomFile *prf,char *arg1,char *arg2,char *arg3)
+    {
+    if(!arg3)
+	{
+	arg3=arg2;
+	arg2=arg1;
+	arg1="file";
+	}
+
+    prf->eType=strcasecmp(arg1,"file") ? RAND_EGD : RAND_FILE;
+    prf->szFile=arg2;
+    prf->nBytes=atoi(arg3);
+    }
+
 static const char *SSLRandomFile(cmd_parms *cmd,char *struct_ptr,char *arg1,
-				 char *arg2)
+				 char *arg2,char *arg3)
     {
     RandomFile rf;
 
-    rf.szFile=arg1;
-    rf.nBytes=atoi(arg2);
+    SetupRandomFile(&rf,arg1,arg2,arg3);
     ReadRandomFiles(cmd->server,&rf,1);
 
     return NULL;
     }
 
 static const char *SSLRandomFilePerConnection(cmd_parms *cmd,char *struct_ptr,
-					      char *arg1,char *arg2)
+					      char *arg1,char *arg2,char *arg3)
     {
     SSLConfigRec *pConfig=ap_get_module_config(cmd->server->module_config,
 					       &apache_ssl_module);
     RandomFile *prf=ap_push_array(pConfig->aRandomFiles);
 
-    prf->szFile=arg1;
-    prf->nBytes=atoi(arg2);
+    SetupRandomFile(prf,arg1,arg2,arg3);
 
     return NULL;
     }
@@ -1926,6 +2159,9 @@ command_rec ssl_cmds[]=
     { "SSLFakeBasicAuth",set_server_bool_slot,
       (void *)XtOffsetOf(SSLConfigRec,bFakeBasicAuth),RSRC_CONF,NO_ARGS,
       "Translate client X509 into a user name" },
+    { "SSLCheckClientDN",set_server_string_slot,
+      (void *)XtOffsetOf(SSLConfigRec,szDNCheckFile),RSRC_CONF,TAKE1,
+      "Check client DN against a list in a file" },
     { "SSLRequiredCiphers",set_server_string_slot,
       (void *)XtOffsetOf(SSLConfigRec,szReqCiphers),RSRC_CONF,TAKE1,
       "Colon-delimited list of required ciphers" },
@@ -1948,16 +2184,20 @@ command_rec ssl_cmds[]=
     { "SSLSessionCacheTimeout",set_server_int_slot,
       (void *)XtOffsetOf(SSLConfigRec,tSessionCacheTimeout),RSRC_CONF,TAKE1,
       "Lifetime in seconds of session cache objects" },
-    { "SSLCacheServerPath",SSLCacheServerPath,NULL,RSRC_CONF,TAKE1,
-      "Path of the global cache server" },
+    { "SSLCacheServerPath",SSLCacheServerPath,NULL,RSRC_CONF,TAKE123,
+      "Path of the global cache server and up to two arguments" },
     { "SSLCacheServerRunDir",SSLCacheServerDir,NULL,RSRC_CONF,TAKE1,
       "Path for the server to run in (in case of coredumps)" },
 #endif
-    { "SSLRandomFile",SSLRandomFile,NULL,RSRC_CONF,TAKE2,
-      "A file to read randomness from and a maximum number of bytes to read" },
+#if ENGINE_SUPPORT
+    { "SSLEngineID",SSLEngineID,NULL,RSRC_CONF,TAKE1,
+      "Alternative OpenSSL `ENGINE id' to use" },
+#endif
+    { "SSLRandomFile",SSLRandomFile,NULL,RSRC_CONF,TAKE23,
+      "A type (file or EGD), a file to read randomness from and a maximum number of bytes to read" },
     { "SSLRandomFilePerConnection", SSLRandomFilePerConnection,NULL,RSRC_CONF,
-      TAKE2,
-      "A file to read randomness from and a maximim number of bytes to read on each request" },
+      TAKE23,
+      "A type (file or EGD), a file to read randomness from and a maximim number of bytes to read on each request" },
     { "SSLLogFile",SSLLogFile,NULL,RSRC_CONF,TAKE1,
       "Defunct SSL log file - DO NOT USE" },
     { NULL }
@@ -1965,7 +2205,7 @@ command_rec ssl_cmds[]=
 
 module MODULE_VAR_EXPORT apache_ssl_module =
     {
-    STANDARD_MODULE_STUFF,
+    SSL_MODULE_STUFF,
     InitSSLServer,		/* initializer */
     CreateSSLDirConfig,		/* dir config creater */
     MergeSSLDirConfig,		/* dir merger --- default is to override */
@@ -1976,9 +2216,15 @@ module MODULE_VAR_EXPORT apache_ssl_module =
     NULL,			/* filename translation */
     FakeBasicAuth,		/* check_user_id */
     NULL,			/* check auth */
-    SSLCheckCipher,		/* check access */
+    CheckAccess,		/* check access */
     NULL,			/* type_checker */
     /*    SSLFixups,			fixups */
     NULL,			/* fixups */
     NULL,			/* logger */
+    NULL,			/* header_parser */
+    NULL,			/* child_init */
+    NULL,			/* child_exit */
+    NULL,			/* post_read_request */
+    SSLSetupConnection2,	/* setup_connection */
+    SSLAddCommonVars		/* add_common_vars */
     };
